"""Constant parameters in support of mirage simulations."""

# NIRISS filter wheel elements
# Source: CSA-JWST-CD-0003: Near Infrared Imager and Slitless Spectrograph Operations Concept Document
NIRISS_PUPIL_WHEEL_ELEMENTS = 'CLEARP F090W F115W F140M F150W F158M F200W NRM GR700XD'.split()
NIRISS_FILTER_WHEEL_ELEMENTS = 'CLEAR GR150R GR150C F277W F356W F380M F430M F444W F480M'.split()