'''Simulate a commissioning PSF using WebbPSF.

*** NOT CURRENTLY IN USE ****
'''

import webbpsf


def create_commissioning_psf(params):
    '''Placeholder function for ultimate WebbPSF query.

    Arguments
    ---------
    params
        Set of parameters that will be used to make a custom PSF with
        WebbPSF (i.e. detector, filter, countrate, sub-pixel position,
        APT #, observation/visit #, etc...)
    '''
	return None